/* Copyright (C) 1998 Free Software Foundation, Inc.

   This file is part of GNU Motti.

   GNU Motti is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Motti is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Motti; see the file COPYING.  If not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stdlib.h>
#include <string.h>

#include "map.h"
#include "misc.h"
#include "snakes.h"

typedef struct
{
  Coord now[2], last[2], start;
  short me, others;
} Snake;

static const Coord bearing_coord[] = {
  {-1, -1},
  {0 , -1},
  {1 , -1},
  {1 , 0 },
  {1 , 1 },
  {0 , 1 },
  {-1, 1 },
  {-1, 0 }};

static int n_snakes = 0;
static Snake *snakes = NULL;

static void free_snake(int);
static int go_on_zeros(Map *, int, Bearing, int);

static void
free_snakes()
{
  free(snakes);
  snakes = NULL;
  n_snakes = 0;
}

static void
add_snake(Map *map, Coord start_loc, Coord last_loc, int players,
	  int others)
{
  if(++others == map->players)
    return;
  snakes = (Snake *) realloc(snakes, sizeof(Snake) * (n_snakes+1));
  snakes[n_snakes].start = snakes[n_snakes].now[0] =
    snakes[n_snakes].now[1] = start_loc;
  snakes[n_snakes].me = players;
  snakes[n_snakes].others = others;
  snakes[n_snakes].last[0] = snakes[n_snakes++].last[1] = last_loc;
}

static Bearing
get_bearing(Map *map, Coord from, Coord to)
{
  register int i;
  Coord j;

  for(i = 0; i < 8 ; i++)
    {
      j = add_coord(from, bearing_coord[i]);
      if(cmp_coord(*map, j, to))
	return i;
    }
  /* This is an impossible condition.  */
  exit(1);
}

static int
check_path(Map *map, int s_n, Bearing b, int direction)
{
  map_val m_val;

  m_val = get_map_val(*map, add_coord(snakes[s_n].now[direction],
				      bearing_coord[b]));
  /* A snake doesn't have to go through an area somebody has already
     passed by.  */
  if(m_val & MASK_PASSED)
    {
      free_snake(s_n);
      return 1;
    }
  /* Check if an area is passable.  */
  if((m_val & MASK_OCCUPIED && (m_val & MASK_PLAYER) == map->op_val)
     || !(m_val & MASK_OCCUPIED))
    {
      snakes[s_n].last[direction] = snakes[s_n].now[direction];
      snakes[s_n].now[direction] =
        add_coord(snakes[s_n].now[direction], bearing_coord[b]);
      set_map_bit(map, snakes[s_n].now[direction], MASK_PASSED);
      return 1;
    }
  return 0;
}

static void
move_snake(int s_n, Coord move_to, int direction)
{
  snakes[s_n].last[direction] = snakes[s_n].now[direction];
  snakes[s_n].now[direction] = move_to;
}
  
static int
check_path_and_player(Map *map, int s_n, Bearing b, int direction)
{
  map_val m_val;
  register short map_bit;
  Coord new_coord;

  new_coord = add_coord(snakes[s_n].now[direction], bearing_coord[b]);
  m_val = get_map_val(*map, new_coord);
  if(m_val == SEA_VAL)
    return 0;
  map_bit = 1<<(m_val & MASK_PLAYER);
  if(!(snakes[s_n].me & map_bit))
    {
      if(m_val & MASK_OCCUPIED)
	{
	  add_snake(map, new_coord, snakes[s_n].now[direction],
		    snakes[s_n].me | map_bit, snakes[s_n].others);
	}
      move_snake(s_n, new_coord, direction);
      return 1;
    }
  else if(m_val & MASK_OCCUPIED)
    {
      return 0;
    }
  else
    {
      move_snake(s_n, new_coord, direction);
      return 1;
    }
}

static void
wriggle(Map *map, int s_n, int direction, int (*function)
	(Map *, int, Bearing, int))
{
  Bearing b;
  int start;

  start = get_bearing(map, snakes[s_n].now[direction],
		      snakes[s_n].last[direction]);

  switch(direction)
    {
    case CLOCKWISE:
      for(b = (start+1) % 8; b != start; b++, b %= 8)
	{
	  if(function(map, s_n, b, direction))
	    return;
	}
      break;
    case C_CLOCKWISE:
      b = start;
      while(b != start)
	{
	  if(--b < 0)
	    b = 7;
	  if(function(map, s_n, b, direction))
	    return;
	}
      break;
    }
  snakes[s_n].last[direction] = snakes[s_n].now[direction];
}

static void
free_snake(int n)
{
  if(--n_snakes == 0)
    {
      free_snakes();
      return;
    }
  if(n != n_snakes)
    {
      memmove(&snakes[n], &snakes[n+1],
	      sizeof(Snake)*(n_snakes-(n+1)));
    }
  snakes = realloc(snakes, sizeof(Snake)*n_snakes);
}

static void
join_snakes(int s_n1, int s_n2, int which)
{
  snakes[s_n1].now[which] = snakes[s_n2].now[which];
  snakes[s_n1].last[which] = snakes[s_n2].last[which];
  free_snake(s_n2);
}

static void
add_region(Region **region, int s_n, int n_regions)
{
  *region = (Region *) realloc(*region, (n_regions+1)*sizeof(Region));
  region[n_regions]->begin = snakes[s_n].start;
  region[n_regions]->players = snakes[s_n].me;
}

static int
snakes_find_areas(Map *map, Region **areas)
{
  register int i, j, k;
  int pair, n_areas = 0;

  while(n_snakes > 1)
    {
      int skip_rest = 0;
      for(i = 0; i < n_snakes && n_snakes > 1; i++, skip_rest = 0)
	{
	  /* Each snake has a pair */
	  for(j = 0; j < 2 && n_snakes > 1 && skip_rest == 0; j++)
	    {
	      pair = abs(j-1);
	      wriggle(map, i, j, &check_path_and_player);
	      for(k = 0; k < n_snakes; k++)
		{
		  if(k == i) continue;
		  /* Check for another snakes' pairs.  There's no need
		     to check for another snakes' other pairs, because
		     they're moving in the same direction.  Different
		     kinds of snakes may not recognize each other.  */
		  if(snakes[i].me == snakes[k].me
		     && cmp_coord(*map, snakes[i].now[j],
				  snakes[k].now[pair]))
		    {
		      join_snakes(i, k, j);
		      if(n_snakes <= 1 && n_areas == 0)
			{
			  free(*areas);
			  *areas = NULL;
			  return 1;
			}
		      skip_rest = 1;
		      break;
		    }
		}
	      if(skip_rest == 1)
		continue;
	      /* Check if a snake has found its pair.  */
	      if(cmp_coord(*map, snakes[i].now[j],
			   snakes[i].last[pair]))
		{
		  add_region(areas, i, n_areas++);
		  free_snake(i);
		}
	    }
	}
    }
  return n_areas;
}

static void
combine_same_areas(Map *map, int n_areas, Region **areas)
{
  register int i, j;
 begin_loop:
  for(i = 0; i < n_areas; i++)
    {
      for(j = 0; j < n_areas; j++)
	{
	  if(i == j)
	    continue;
	  if(cmp_coord(*map, areas[i]->begin, areas[j]->begin))
	    {
	      areas[i]->players |= areas[j]->players;
	      memmove(areas[j], areas[j+1], n_areas---j);
	      goto begin_loop;
	    }
	}
    }
}

extern int
count_areas(Map *map, Coord loc, Region **areas_start)
{
  unsigned register int i;
  Coord sloc;

  *areas_start = NULL;
  sloc = loc;

  for(i = 0; i < 8; i++)
    {
      map_val test_val;
      sloc = add_coord(loc, bearing_coord[i]);
      test_val = get_map_val(*map, sloc) & MASK_PLAIN;
      if(test_val != (map->turn | MASK_OCCUPIED) && test_val != SEA_VAL)
	{
	  add_snake(map, sloc, loc, 1<<map->turn, 0);
	}
    }
  i = snakes_find_areas(map, areas_start);
  combine_same_areas(map, i, areas_start);
  free_snakes();
  return i;
}

extern map_val
get_surrounding_val(Map *map, Coord loc, Bearing b)
{
  return get_map_val(*map, add_coord(loc, bearing_coord[b]));
}

/* Here are the routines to find capitals.  */

#define HORIZONTAL 0
#define VERTICAL 1

typedef struct
{
  Bearing dir;
  Coord loc;
} Drone;

static int n_drones = 0;
static Drone *drones = NULL;

static void
free_drones()
{
  free(drones);
  drones = NULL;
  n_snakes = 0;
}

static void
add_drone(Coord loc, Bearing dir)
{
  drones = (Drone *) realloc(drones, (n_drones+1)*sizeof(Drone));
  drones[n_drones].loc = loc;
  drones[n_drones++].dir = dir;
}

static void
kill_drone(int n)
{
  if(--n_drones == 0)
    {
      free_drones();
      return;
    }
  if(n != n_drones)
    {
      memmove(&drones[n], &drones[n+1],
	      sizeof(Drone)*(n_drones-(n+1)));
    }
  drones = realloc(drones, sizeof(Drone)*n_drones);
}

static void
check_add_drone(Map *map, Coord loc, int dir, int player)
{
  Coord new_coord;
  map_val new_m_val;
  register int i, j;

  if(dir == HORIZONTAL)
    {
      i = 0;
      j = 2;
    }
  else
    {
      i = 2;
      j = 4;
    }
  for(; i < j; i++)
    {
      Bearing b;
      switch (i)
	{
	case 0:
	  b = E;
	  break;
	case 1:
	  b = W;
	  break;
	case 2:
	  b = N;
	  break;
	case 3:
	  b = S;
	}
      new_coord = add_coord(loc, bearing_coord[b]);
      new_m_val = get_map_val(*map, new_coord);
      /* Don't add a drone if there's an obstacle.  */
      if(new_m_val ^ MASK_PASSED && new_m_val != SEA_VAL
	 && ((new_m_val & MASK_OCCUPIED
	      && (new_m_val & MASK_PLAYER) == player)
	     || new_m_val ^ MASK_OCCUPIED))
	{			
	  add_drone(new_coord, b);
	}
    }
}

extern int
find_capital(Map *map, Coord start, int player)
{
  if(cmp_coord(*map, start, map->capital[player]))
    {
      return 1;
    }
  else
    {
      Coord aux_axis, x_axis, y_axis;
      int dir;
      Coord diff;
      map_val my_occ_val;

      map->op_val = player;
      my_occ_val = player | MASK_OCCUPIED;
      diff.x = map->capital[player].x - start.x;
      diff.y = map->capital[player].y - start.y;
      if(abs(diff.x) < abs(diff.y))
	{
	  dir = VERTICAL;
	  add_drone(start, N);
	  add_drone(start, S);
	}
      else
	{
	  dir = HORIZONTAL;
	  add_drone(start, E);
	  add_drone(start, W);
	}
      aux_axis = start;
      set_map_bit(map, start, MASK_PASSED);
      x_axis.y = map->capital[player].y;
      y_axis.x = map->capital[player].x;
      while(n_drones > 0 && n_snakes > 0)
	{
	  Coord loc;
	  register int i;
	  for(i = 0; i < n_drones; i++)
	    {
	      map_val m_val;
	      loc = add_coord(drones[i].loc,
			      bearing_coord[drones[i].dir]);
	      m_val = get_map_val(*map, loc);
	      x_axis.x = loc.x;
	      y_axis.y = loc.y;
	      /* Check for capital.  */
	      if(cmp_coord(*map, loc, map->capital[player]))
		{
		  free_snakes();
		  free_drones();
		  clean_map(map, MASK_PASSED);
		  return 1;
		}
	      /* Check if a map cell is an obstacle.  */
	      if((m_val & MASK_OCCUPIED && m_val != my_occ_val)
		 || m_val == SEA_VAL)
		{
		  add_snake(map, drones[i].loc, loc, 1<<map->op_val, 0);
		  kill_drone(i);
		  continue;
		}
	      /* Check if this map cell has already been visited.  */
	      else if(m_val & MASK_PASSED)
		{
		  kill_drone(i);
		  continue;
		}
	      /* The drone goes on.  */
	      else
		{
		  drones[i].loc = loc;
		  set_map_bit(map, loc, MASK_PASSED);
		}
	      /* Check if the drone travelling the auxilary axis has
		 hit either of the main axes.  */
	      if(cmp_coord(*map, drones[i].loc, x_axis) &&
		 (drones[i].dir == N || drones[i].dir == S))
		{
		  check_add_drone(map, loc, HORIZONTAL, player);
		}
	      else if(cmp_coord(*map, drones[i].loc, y_axis) &&
		      (drones[i].dir == E || drones[i].dir == W))
		{
		  check_add_drone(map, loc, VERTICAL, player);
		}
	    }
	  for(i = 0; i < n_snakes; i++)
	    {
	      register int j;
	      for(j = 0; j < 2; j++)
		{
		  unsigned register char pair;
		  register int k;
		  pair = abs(j-1);
		  wriggle(map, i, j, &check_path);
		  for(k = 0; k < n_snakes; k++)
		    {
		      if(k == i)
			continue;
		      if(cmp_coord(*map, snakes[i].now[j],
				   snakes[k].now[pair]))
			{
			  join_snakes(i, k, j);
			  break;
			}
		    }
		  if(dir == VERTICAL)
		    {
		      aux_axis.y = snakes[i].now[j].y;
		    }
		  else
		    {
		      aux_axis.x = snakes[i].now[j].x;
		    }
		  x_axis.x = snakes[i].now[j].x;
		  y_axis.y = snakes[i].now[j].y;
		  if(cmp_coord(*map, map->capital[player],
			       snakes[i].now[j]))
		    {
		      free_snakes();
		      free_drones();
		      clean_map(map, MASK_PASSED);
		      return 1;
		    }
		  if(cmp_coord(*map, x_axis, snakes[i].now[j])
		     && get_map_val(*map, x_axis) ^ MASK_PASSED)
		    {
		      check_add_drone(map, x_axis, HORIZONTAL,
				       player);
		    }
		  else if(cmp_coord(*map, y_axis, snakes[i].now[j])
			  && get_map_val(*map, y_axis) ^ MASK_PASSED)
		    {
		      check_add_drone(map, x_axis, VERTICAL, player);
		    }
		  if(cmp_coord(*map, snakes[i].now[j],
			       snakes[i].last[pair]))
		    {
		      free_snake(i);
		    }
		}
	    }
	}
    }
  clean_map(map, MASK_PASSED);
  return 0;
}

#define CLEAR 0
#define PASSED 1
#define BORDER 2
#define MASK_FILL_PASSED 4
#define MASK_UNPASSABLE 6

/* Recursion.  I feel almost like a lisp programmer.  */

static int
find_new_area(Map *map, Coord *new_loc, Coord *last_loc)
{
  register Bearing i;
  Coord new_loc2, last_loc2;
  if(!(get_map_val(*map, *new_loc) & MASK_UNPASSABLE))
    return 1;
  last_loc2 = *new_loc;
  set_map_bit(map, last_loc2, MASK_FILL_PASSED);
  for(i = 0; i < 8; i++)
    {
      new_loc2 = add_coord(last_loc2, bearing_coord[i]);
      if(find_new_area(map, &new_loc2, &last_loc2))
	{
	  clear_map_bit(map, last_loc2, MASK_FILL_PASSED);
	  *new_loc = new_loc2;
	  *last_loc = last_loc2;
	  return 1;
	}
    }
  return 0;
}

static int
go_on_zeros(Map *map, int s_n, Bearing b, int direction)
{
  Coord new_coord;
  new_coord = add_coord(snakes[s_n].now[direction], bearing_coord[b]);
  if(get_map_val(*map, new_coord) == 0)
    {
      snakes[s_n].last[direction] = snakes[s_n].now[direction];
      snakes[s_n].now[direction] = new_coord;
      return 1;
    }
  return 0;
}

static int
occupy(Map *map, Coord loc, int players, Coord **event_loc)
{
  int p_array[MAXPLAYERS] = {0};
  register Bearing i;
  int greatest_val = 0, greatest_pl;
  for(i = 0; i < 8; i++)
    {
      Coord sur_coord;
      register int sur_coord_pl;
      sur_coord = add_coord(loc, bearing_coord[i]);
      sur_coord_pl = (get_map_val(*map, sur_coord) & MASK_PLAYER);
      if(1<<sur_coord_pl & players)
	p_array[sur_coord_pl]++;
    }
  for(i = 1; i <= map->players; i++)
    {
      if(p_array[i] > greatest_val)
	{
	  greatest_val = p_array[i];
	  greatest_pl = i;
	}
    }
  if(greatest_val != 0)
    {
      set_map(map, loc, greatest_pl);
      **(event_loc++) = loc;
      return 1;
    }
  else
    return 0;
}
 
extern void
encircled(Map *map, Coord last_loc, Coord loc, int players, Action *act)
{
  Map map_copy;
  register int i;
  Coord *event_loc_ptr;

  /* Definitely enough space for all the occupy events!  */
  act->loc = (Coord *) realloc(act->loc, (act->base_count+map->size) *
			       sizeof(Coord));
  event_loc_ptr = act->loc+act->base_count;
  map_copy.width = map->width;
  map_copy.height = map->height;
  map_copy.warped = map->warped;
  map_copy.map = (map_val *) mymalloc(map->size*sizeof(map_val));
  for(i = 0; i < map->size; i++)
    {
      if((map->map[i] & MASK_OCCUPIED)
	 && 1<<(map->map[i] & MASK_PLAYER) & players)
	map_copy.map[i] = BORDER;
      else
	map_copy.map[i] = CLEAR;
    }
  while(find_new_area(&map_copy, &loc, &last_loc))
    {
      add_snake(map, loc, last_loc, 0, 0);
      set_map_bit(&map_copy, loc, PASSED);
      if(!(1<<(get_map_val(*map, loc) & MASK_PLAYER) & players))
	occupy(map, loc, players, &event_loc_ptr);
      while(n_snakes == 1)
	{
	  register int i;
	  for(i = 0; i < 2; i++)
	    {
	      wriggle(&map_copy, 0, i, &go_on_zeros);
	      set_map_bit(&map_copy, snakes[0].now[i], PASSED);
	      if(cmp_coord(*map, snakes[0].now[i],
			   snakes[0].last[abs(i-1)]))
		{
		  free_snake(0);
		  break;
		}
	    }
	}
    }
  free(map_copy.map);
}

extern void
eliminate(Map *map, Action *act, int player)
{
  int areas_left = 1;
  int player_bits;
  Coord *event_loc_ptr;
  player_bits = ~(1<<player);
  act->loc = (Coord *) realloc(act->loc, (act->base_count+map->size) *
			       sizeof(Coord));
  event_loc_ptr = act->loc+act->base_count;
  while(areas_left)
    {
      register Coord loc;
      areas_left = 0;
      for(loc.x = 0; loc.x < map->width; loc.x++)
	{
	  for(loc.y = 0; loc.y < map->height; loc.y++)
	    {
	      if((get_map_val(*map, loc) & MASK_PLAYER) == player
		 && occupy(map, loc, player_bits, &event_loc_ptr))
		areas_left = 1;
	    }
	}
    }
}
